from dataclasses import dataclass, field
from decimal import Decimal
from typing import Iterable, Literal

from .base import AbacusError, Numeric, Posting


@dataclass
class Debit(Posting):
    """Increase debit-normal accounts, decrease credit-normal accounts."""

    account: str
    amount: int | float | Decimal
    tag: Literal["debit"] = "debit"


@dataclass
class Credit(Posting):
    """Increase credit-normal accounts, decrease debit-normal accounts."""

    account: str
    amount: int | float | Decimal
    tag: Literal["credit"] = "credit"


@dataclass
class Double(Posting, Iterable[Debit | Credit]):
    """Double-entry transaction."""

    debit: str
    credit: str
    amount: int | float | Decimal
    tag: Literal["double"] = "double"

    def __iter__(self):
        yield Debit(self.debit, self.amount)
        yield Credit(self.credit, self.amount)


@dataclass
class Unbalanced(Posting, Iterable[Debit | Credit]):
    """Multiple entry that is not guaranteed to be balanced by debits and credits."""

    debits: list[tuple[str, Numeric]] = field(default_factory=list)
    credits: list[tuple[str, Numeric]] = field(default_factory=list)

    def __iter__(self):
        for account, amount in self.debits:
            yield Debit(account, Decimal(amount))
        for account, amount in self.credits:
            yield Credit(account, Decimal(amount))

    @staticmethod
    def sums(xs):
        return sum(x for (_, x) in xs)

    def is_balanced(self):
        return self.sums(self.debits) == self.sums(self.credits)

    def to_multiple(self):
        return Multiple(debits=self.debits, credits=self.credits)


@dataclass
class Multiple(Unbalanced):
    """Multiple entry that is balanced by debits and credits."""

    tag: Literal["multiple"] = "multiple"

    @classmethod
    def from_list(cls, singles):
        entry = Unbalanced()
        for s in singles:
            match s:
                case Debit(account, amount):
                    entry.debits.append((account, amount))
                case Credit(account, amount):
                    entry.credits.append((account, amount))
        return entry.to_multiple()

    def __post_init__(self):
        self.validate()

    def validate(self):
        ds = self.sums(self.debits)
        cs = self.sums(self.credits)
        if ds != cs:
            raise AbacusError(
                f"Debits {ds} and credits {cs} are not balanced for {self}."
            )
        return self


@dataclass
class Initial(Posting):
    """Open ledger with initial balances."""

    balances: dict[str, Numeric]
    tag: Literal["initial"] = "initial"


@dataclass
class Entry:
    title: str
    debits: list[tuple[str, Decimal]] = field(default_factory=list)
    credits: list[tuple[str, Decimal]] = field(default_factory=list)
    _amount: Decimal | None = None

    def amount(self, amount):
        """Set amount for the entry."""
        self._amount = Decimal(amount)
        return self

    def get_amount(self, amount: Numeric | None = None) -> Decimal:
        """Use provided amount, default amount or raise error if no data about amount."""
        if amount is None:
            if self._amount:
                return self._amount
            else:
                raise AbacusError("Amount is not set.")
        else:
            return Decimal(amount)

    def debit(self, account, amount=None):
        amount = self.get_amount(amount)
        self.debits.append((account, amount))
        return self

    def credit(self, account, amount=None):
        amount = self.get_amount(amount)
        self.credits.append((account, amount))
        return self

    def double(self, debit, credit, amount):
        self.debit(debit, amount)
        self.credit(credit, amount)
        return self

    def to_multiple(self):
        return Multiple(debits=self.debits, credits=self.credits)  # type: ignore

    def __iter__(self):
        return iter(self.to_multiple())



from .base import T5, AbacusError
from .book import Book
from .chart import Add, Asset, Chart, Equity, Expense, Income, Liability, Offset
from .entry import Credit, Debit, Double, Entry, Initial, Multiple
from .ledger import Close, Contra, Event, History, Ledger



"""Chart of accounts and events to modify it.

The primitive events are:

- `Add` and `Offset` to add regular and contra accounts,
- `Drop` to deactivate an empty account.

`Account` is a parent class
for `Asset`, `Equity`, `Liability`, `Income`, and `Expense`.
`Account` is a compound event
that translates to a sequence of `Add` and `Offset` events.

`BaseChart` contains account names and titles.
A list of `Account` objects is enough to define a `BaseChart` object.

`Earnings` class indicates current and retained earnings account names
that we need for closing the ledger at period end.

`Chart` is a serializable chart of accounts that:
-  can be saved and loaded from a flat and readable JSON,
-  has 4 validation methods for consistency checks,
-  contains `BaseChart` and `Earnings` objects.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Iterable, Iterator, Literal

from pydantic import BaseModel, ConfigDict

from .base import T5, AbacusError, Charting, Operation, SaveLoadMixin


@dataclass
class Add(Charting):
    """Add account."""

    name: str
    t: T5
    tag: Literal["add"] = "add"


@dataclass
class Offset(Charting):
    """Add contra account."""

    parent: str
    name: str
    tag: Literal["offset"] = "offset"


@dataclass
class Drop(Charting):
    """Drop account if the account and its contra accounts have zero balances."""

    name: str
    tag: Literal["drop"] = "drop"


@dataclass
class Account(ABC, Operation, Iterable[Add | Offset]):
    name: str
    contra_accounts: list[str] = field(default_factory=list)
    title: str | None = None

    @property
    def headline(self) -> str:
        return f"{self.tag}:{self.name}"

    @property
    @abstractmethod
    def tag(self):
        pass

    @property
    def t(self) -> T5:
        return T5(self.tag)

    def __iter__(self) -> Iterator[Add | Offset]:
        yield Add(self.name, self.t)
        for contra_name in self.contra_accounts:
            yield Offset(self.name, contra_name)


@dataclass
class Asset(Account):
    tag: Literal["asset"] = "asset"  # type: ignore


@dataclass
class Equity(Account):
    tag: Literal["equity"] = "equity"  # type: ignore


@dataclass
class Liability(Account):
    tag: Literal["liability"] = "liability"  # type: ignore


@dataclass
class Income(Account):
    tag: Literal["income"] = "income"  # type: ignore


@dataclass
class Expense(Account):
    tag: Literal["expense"] = "expense"  # type: ignore


class BaseChart(BaseModel):
    """Chart of accounts without earnings accounts."""

    model_config = ConfigDict(extra="forbid")

    assets: list[str] = []
    equity: list[str] = []
    liabilities: list[str] = []
    income: list[str] = []
    expenses: list[str] = []
    contra_accounts: dict[str, list[str]] = {}
    names: dict[str, str] = {}

    def extend(self, accounts: Iterable["Account"]):
        """Add a list of accounts to chart."""
        for account in accounts:
            self.add_account(account)
        return self

    def add_account(self, account: "Account"):
        """Add account to chart."""
        self.add_string(account.headline)
        for contra in account.contra_accounts:
            self.offset(account.name, contra)
        if account.title:
            self.name(account.name, account.title)
        return self

    def add_string(self, string: str):
        """Add account by slug like `asset:cash` or `a:cash`."""
        prefix, account_name = string.split(":")
        match prefix[0].lower():
            case "a":
                self.assets.append(account_name)
            case "l":
                self.liabilities.append(account_name)
            case "c":
                self.equity.append(account_name)
            case "e":
                match prefix[1].lower():
                    case "q":
                        self.equity.append(account_name)
                    case "x":
                        self.expenses.append(account_name)
                    case _:
                        raise AbacusError(f"Invalid prefix: {prefix}")
            case "i":
                self.income.append(account_name)
            case _:
                raise AbacusError(f"Invalid prefix: {prefix}")
        return self

    def offset(self, account_name: str, contra_name: str):
        """Add contra account to chart."""
        AbacusError.must_exist(self.account_names, account_name)
        if contra_name in self.contra_accounts:
            self.contra_accounts[account_name].append(contra_name)
        else:
            self.contra_accounts[account_name] = [contra_name]
        return self

    def name(self, account_name: str, title: str):
        """Add descriptive account title."""
        self.names[account_name] = title
        return self

    @property
    def matching(self):
        """Match attributes with account classes."""
        return (
            (Asset, "assets"),
            (Equity, "equity"),
            (Liability, "liabilities"),
            (Income, "income"),
            (Expense, "expenses"),
        )

    @property
    def accounts(self) -> list["Account"]:
        """Accounts from this chart."""
        accounts = []
        for cls, attribute in self.matching:
            for account_name in getattr(self, attribute):
                contra_names = self.contra_accounts.get(account_name, [])
                title = self.names.get(account_name, None)
                accounts.append(cls(account_name, contra_names, title))
        return accounts

    @property
    def regular_names(self):
        """All regular account names."""
        return [add.name for add in self.accounts]

    @property
    def contra_names(self) -> list[str]:
        """All contra account names."""
        return [
            name
            for contra_names in self.contra_accounts.values()
            for name in contra_names
        ]

    @property
    def account_names(self) -> list[str]:
        """All accounts in this chart including the duplicates."""
        return self.regular_names + self.contra_names

    @property
    def duplicates(self) -> list[str]:
        """Duplicate account names. Must be empty for valid chart."""
        names = self.account_names
        for name in set(names):
            names.remove(name)
        return names

    def assert_account_names_are_unique(self):
        """Raise error if duplicate account names are found."""
        if ds := self.duplicates:
            raise AbacusError(f"Account names are not unique: {ds}")

    def assert_contra_account_references_are_valid(self):
        """Raise error if any contra account reference is invalid."""
        regular_names = self.regular_names
        for parent_account in self.contra_accounts.keys():
            AbacusError.must_exist(regular_names, parent_account)

    def to_chart(self, current_earnings: str, retained_earnings: str) -> "Chart":
        """Convert to chart using current and retained earnings account names."""
        return Chart(
            **self.model_dump(),
            current_earnings=current_earnings,
            retained_earnings=retained_earnings,
        )


class Chart(BaseChart, SaveLoadMixin):
    """Serializable chart of accounts that is saved to plain JSON."""

    current_earnings: str
    retained_earnings: str

    @classmethod
    def new(
        cls,
        current_earnings: str,
        retained_earnings: str,
        accounts: Iterable["Account"],
    ):
        """Create chart from account classes."""
        self = cls(
            current_earnings=current_earnings, retained_earnings=retained_earnings
        )
        self.extend(accounts)
        return cls(**self.model_dump())

    def model_post_init(self, _):
        self.move_retained_earnings_to_chart()
        self.assert_current_earnings_not_in_chart()
        self.assert_account_names_are_unique()
        self.assert_contra_account_references_are_valid()

    def move_retained_earnings_to_chart(self):
        """Add retained earnings to equity account if not already there."""
        re = Equity(self.retained_earnings)
        if re.name not in self.equity:
            self.add_account(re)

    def assert_current_earnings_not_in_chart(self):
        """Raise error if current earnings is in chart."""
        AbacusError.must_not_exist(self.account_names, self.current_earnings)

    @property
    def base(self) -> BaseChart:
        """Base chart without earnings."""
        dump = self.model_dump()
        del dump["current_earnings"]
        del dump["retained_earnings"]
        return BaseChart(**dump)

    @property
    def earnings(self) -> "Earnings":
        """Earnings account names from this chart."""
        return Earnings(current=self.current_earnings, retained=self.retained_earnings)

    # @property
    # def qualified(self) -> "QualifiedChart":
    #     return QualifiedChart(earnings=self.earnings, base=self.base)


class Earnings(BaseModel):
    current: str
    retained: str

    def to_chart(self, accounts: Iterable["Account"]) -> Chart:
        return BaseChart().extend(accounts).to_chart(self.current, self.retained)


# @dataclass
# class QualifiedChart:
#     earnings: Earnings
#     base: BaseChart

#     def __post_init__(self):
#         self.to_chart()

#     def __iter__(self) -> Iterator["Account"]:
#         return iter(self.base.accounts)

#     def to_chart(self) -> Chart:
#         return self.earnings.to_chart(self.base.accounts)



"""An accounting ledger that equates to a sequence of events.

The main class is `Ledger`. You can modify the state of ledger by applying operations to it.

The basic ('primitive') operations are:
- `Add` and `Offset` to add regular and contra accounts,
- `Debit` and `Credit` to change account balances,
- `PeriodEnd` to mark accounting period end,
- `Drop` to deactivate an empty account.

The compound operations are:
- `Account` to specify an account together with contra accounts,
- `Initial` to open ledger with initial balances,
- `Double` to make a double entry,
- `Multiple` to make a multiple entry,
- `Transfer` to move account balance to another account,
- `Close` for closing accounts at period end.

Each compound operation consists of a sequence of basic operations.

`Ledger.history` holds a sequence of events that where applied to the ledger.
You can re-run the events on empty ledger and will arrive to the same state of ledger.
"""

from abc import ABC, abstractmethod
from collections import UserDict
from copy import deepcopy
from dataclasses import dataclass, field
from decimal import Decimal
from typing import Iterable, Literal

import simplejson as json  # type: ignore
from pydantic import BaseModel

from .base import T5, AbacusError, Closing, Numeric, Operation, SaveLoadMixin
from .chart import Account, Add, Asset, Drop, Equity, Expense, Income, Liability, Offset
from .entry import Credit, Debit, Double, Initial, Multiple, Unbalanced


@dataclass
class Transfer(Closing):
    """Transfer account balance to another account."""

    from_account: str
    to_account: str
    tag: Literal["transfer"] = "transfer"


@dataclass
class PeriodEnd(Closing):
    """Mark end of accounting period and save the copy of ledger
    to be used for the income statement.
    """

    tag: Literal["period_end"] = "period_end"


@dataclass
class Close(Closing):
    """Close ledger to earnings account."""

    earnings_account: str
    tag: Literal["close"] = "close"


@dataclass
class TAccount(ABC):
    balance: Decimal = Decimal(0)

    @abstractmethod
    def debit(self, amount: int | float | Decimal):
        pass

    def credit(self, amount: int | float | Decimal):
        self.debit(-amount)

    def is_empty(self) -> bool:
        return self.balance == Decimal(0)

    @abstractmethod
    def transfer(self, from_account: str, to_account: str) -> Double:
        pass


class DebitAccount(TAccount):
    def debit(self, amount: int | float | Decimal):
        self.balance += Decimal(amount)

    def transfer(self, from_account: str, to_account: str) -> Double:
        return Double(to_account, from_account, self.balance)


class CreditAccount(TAccount):
    def debit(self, amount: int | float | Decimal):
        self.balance -= Decimal(amount)

    def transfer(self, from_account: str, to_account: str) -> Double:
        return Double(from_account, to_account, self.balance)


@dataclass
class Contra:
    """Contra account, refers to an existing regular account."""

    name: str


class ChartDict(UserDict[str, T5 | Contra]):
    def is_debit_account(self, name) -> bool:
        AbacusError.must_exist(self, name)
        match self[name]:
            case Contra(name):
                return not self.is_debit_account(name)
            case t:
                return t in {T5.Asset, T5.Expense}

    def by_type(self, t: T5) -> list[str]:
        return [name for name, account_type in self.data.items() if account_type == t]

    def find_contra_accounts(self, regular_account_name: str) -> list[str]:
        return [
            contra_name
            for contra_name, parent in self.data.items()
            if parent == Contra(regular_account_name)
        ]

    def close_contra_accounts(self, t: T5) -> Iterable[tuple[str, str]]:
        for account_name in self.by_type(t):
            for contra_name in self.find_contra_accounts(account_name):
                yield (contra_name, account_name)

    def close_type(self, t: T5, earnings_account: str) -> Iterable[tuple[str, str]]:
        for account_name in self.by_type(t):
            yield (account_name, earnings_account)

    def close(self, earnings_account: str) -> Iterable[tuple[str, str]]:
        for t in (T5.Income, T5.Expense):
            yield from self.close_contra_accounts(t)
            yield from self.close_type(t, earnings_account)

    def initial_entry(self, balances: dict[str, Numeric]) -> Multiple:
        entry = Unbalanced()
        for account, amount in balances.items():
            AbacusError.must_exist(self, account)
            pair = account, Decimal(amount)
            if self.is_debit_account(account):
                entry.debits.append(pair)
            else:
                entry.credits.append(pair)
        return entry.to_multiple()


Primitive = Add | Offset | Debit | Credit | PeriodEnd | Drop
AccountType = Asset | Equity | Liability | Income | Expense
EntryType = Double | Multiple | Initial
ClosingType = Transfer | Close
# may disqualify primitives from actions or allow a list of primitives in actions
Action = Primitive | AccountType | EntryType | ClosingType


@dataclass
class Event:
    action: Action
    primitives: list[Primitive]
    note: str | None


class History(BaseModel, SaveLoadMixin):
    events: list[Event] = field(default_factory=list)

    def __iter__(self):  # type: ignore
        """Override `pydantic.BaseModel.__iter__`."""
        yield from self.events

    def append(
        self, action: Action, primitives: list[Primitive], note: str | None = None
    ):
        event = Event(action, primitives, note)
        self.events.append(event)

    @property
    def primitives(self) -> Iterable[Primitive]:
        for event in self.events:
            yield from event.primitives

    @property
    def actions(self) -> Iterable:
        for event in self.events:
            yield event.action

    @property
    def accounts(self) -> Iterable[Account]:
        for action in self.actions:
            if isinstance(action, Account):
                yield action

    def to_ledger(self) -> "Ledger":
        """Re-create ledger from history of actions."""
        return Ledger.from_list(self.actions)


@dataclass
class Ledger:
    accounts: dict[str, TAccount] = field(default_factory=dict)
    chart: ChartDict = field(default_factory=ChartDict)
    history: History = field(default_factory=History)
    accounts_before_close: dict[str, TAccount] | None = None

    def is_closed(self) -> bool:
        return self.accounts_before_close is not None

    @classmethod
    def from_accounts(cls, accounts: Iterable[Account]):
        return cls.from_list(accounts)  # type: ignore

    @classmethod
    def from_list(cls, actions: Iterable[Action]):
        return cls().apply_many(actions)

    @property
    def balances(self):
        return ReportDict(
            {name: account.balance for name, account in self.accounts.items()}
        )

    def create_account(self, name):
        if self.chart.is_debit_account(name):
            self.accounts[name] = DebitAccount()
        else:
            self.accounts[name] = CreditAccount()

    def run_iterable(self, action: Account | Double | Multiple) -> list[Primitive]:
        """Get primitives from an iterable action and use them to change ledger."""
        primitives: list[Primitive] = []
        for operation in iter(action):
            primitives += self.run(operation)
        return primitives

    def run(self, action: Operation) -> list[Primitive]:
        """Apply action and return list of primitives."""
        if isinstance(action, (Account, Double, Multiple)):
            return self.run_iterable(action)
        operations: list[Primitive] = []
        match action:
            case Add(name, t):
                AbacusError.must_not_exist(self.chart, name)
                self.chart[name] = t
                self.create_account(name)
                return [action]
            case Offset(parent, name):
                if parent not in self.chart:
                    raise AbacusError(f"Account {parent} must exist.")
                AbacusError.must_not_exist(self.chart, name)
                self.chart[name] = Contra(parent)
                self.create_account(name)
                return [action]
            case Drop(name):
                if not self.accounts[name].is_empty():
                    raise AbacusError(f"Account {name} is not empty.")
                    # also check contra accounts are empty
                del self.accounts[name]
                return [action]
            case Debit(account, amount):
                AbacusError.must_exist(self.accounts, account)
                self.accounts[account].debit(amount)
                return [action]
            case Credit(account, amount):
                AbacusError.must_exist(self.accounts, account)
                self.accounts[account].credit(amount)
                return [action]
            case Initial(balances):
                initial_entry = self.chart.initial_entry(balances)
                return self.run(initial_entry)
            case Transfer(from_account, to_account):
                account_ = self.accounts[from_account]
                transfer_entry = account_.transfer(from_account, to_account)
                operations += self.run(transfer_entry)
                return operations
            case PeriodEnd():
                self.accounts_before_close = deepcopy(self.accounts)
                return [action]
            case Close(earnings_account):
                for operation in self.close_ledger_items(earnings_account):
                    operations += self.run(operation)
                return operations
            case _:
                raise AbacusError(f"Unknown {action}")

    def close_ledger_items(self, earnings_account: str) -> Iterable[Operation]:
        yield PeriodEnd()
        for from_account, to_account in self.chart.close(earnings_account):
            yield Transfer(from_account, to_account)
            yield Drop(from_account)

    def close(self, earnings_account: str):
        self.apply(Close(earnings_account))
        return self

    def apply(self, action: Action, note: str | None = None):
        operations = self.run(action)
        self.history.append(action, operations, note)
        return self

    def apply_many(self, actions: Iterable[Action], note: str | None = None):
        for action in actions:
            self.apply(action, note)
        return self

    def income_statement(self) -> "IncomeStatement":
        if self.is_closed():
            return IncomeStatement.new(self.accounts_before_close, self.chart)
        else:
            return IncomeStatement.new(self.accounts, self.chart)

    def proxy(self, proxy_earnings_account: str) -> "Ledger":
        """Create a shallow ledger copy."""
        return (
            Ledger(accounts=deepcopy(self.accounts), chart=deepcopy(self.chart))
            .apply(Equity(proxy_earnings_account))
            .close(proxy_earnings_account)
        )

    def balance_sheet(self, proxy_earnings: str = "current_earnings") -> "BalanceSheet":
        ledger = self if self.is_closed() else self.proxy(proxy_earnings)
        return BalanceSheet.new(ledger.accounts, ledger.chart)


@dataclass
class Reporter:
    accounts: dict[str, TAccount]
    chart: ChartDict

    def net_balance(self, name: str) -> Decimal:
        """Return account balance minus associated contra account balances."""
        contra_account_balances = [
            self.accounts[contra_name].balance
            for contra_name in self.chart.find_contra_accounts(name)
        ]
        return Decimal(self.accounts[name].balance - sum(contra_account_balances))

    def fill(self, t: T5) -> "ReportDict":
        """Return net balances for a given account type."""
        result = ReportDict()
        for name in self.chart.by_type(t):
            result[name] = self.net_balance(name)
        return result


class ReportDict(UserDict[str, Decimal], SaveLoadMixin):
    @property
    def total(self):
        return Decimal(sum(self.data.values()))

    def model_dump_json(self, indent: int = 2, warnings: bool = False):
        return json.dumps(self.data, indent=indent)

    @classmethod
    def model_validate_json(cls, text: str):
        return cls(json.loads(text))


class Report:
    """Base class for financial reports."""


@dataclass
class IncomeStatement(Report):
    income: ReportDict
    expenses: ReportDict

    @classmethod
    def new(cls, balances, chart):
        reporter = Reporter(balances, chart)
        return cls(income=reporter.fill(T5.Income), expenses=reporter.fill(T5.Expense))

    @property
    def net_earnings(self):
        """Calculate net earnings as income less expenses."""
        return self.income.total - self.expenses.total


@dataclass
class BalanceSheet(Report):
    assets: ReportDict
    equity: ReportDict
    liabilities: ReportDict

    @classmethod
    def new(cls, balances, chart):
        reporter = Reporter(balances, chart)
        return cls(
            assets=reporter.fill(T5.Asset),
            equity=reporter.fill(T5.Equity),
            liabilities=reporter.fill(T5.Liability),
        )

    def is_balanced(self) -> bool:
        """Return True if assets equal liabilities plus equity."""
        return self.assets.total == (self.equity.total + self.liabilities.total)



"""User-facing Book class for an accounting ledger."""

from dataclasses import dataclass
from typing import Iterable

from .chart import Chart, Earnings
from .entry import Entry
from .ledger import History, Initial, Ledger


@dataclass
class Book:
    earnings: Earnings
    ledger: Ledger

    @classmethod
    def from_chart(cls, chart: Chart):
        ledger = Ledger.from_accounts(chart.accounts)
        return cls(chart.earnings, ledger)

    @property
    def chart(self) -> Chart:
        return self.earnings.to_chart(self.ledger.history.accounts)

    def open(self, balances: dict):
        self.ledger.apply(Initial(balances))

    def post(self, entry: Entry):
        self.ledger.apply(entry.to_multiple(), entry.title)

    def post_many(self, entries: Iterable[Entry]):
        for entry in entries:
            self.post(entry)

    def close(self):
        self.ledger.close(self.earnings.retained)

    @property
    def balances(self):
        return self.ledger.balances

    @property
    def income_statement(self):
        return self.ledger.income_statement()

    @property
    def balance_sheet(self):
        return self.ledger.balance_sheet(self.earnings.current)

    def save_chart(self, path, allow_overwrite=False):
        self.chart.save(path, allow_overwrite)

    @classmethod
    def from_chart_dump(cls, path):
        chart = Chart.load(path)
        return cls(chart.earnings, Ledger())

    def save_history(self, path, allow_overwrite=False):
        self.ledger.history.save(path, allow_overwrite)

    def load_history(self, path):
        return History.load(path)

    def save(self, chart_path, history_path, allow_overwrite=False):
        """Save book to chart and history JSON files."""
        self.chart.save(chart_path, allow_overwrite)
        self.ledger.history.save(history_path, allow_overwrite)

    @classmethod
    def load_unsafe(cls, chart_path, history_path):
        """Unsafe method to load a Book from chart and history JSON files."""
        self = cls.from_chart_dump(chart_path)
        history = self.load_history(history_path)
        self.ledger = Ledger.from_accounts(history.accounts)
        # pray that the history is consistent with the chart (fixme)
        return self



from decimal import Decimal
from enum import Enum
from pathlib import Path
from typing import Iterable

Numeric = int | float | Decimal


class Operation(object):
    """A unit of change of the ledger state.

    Types of operations:
    - Account and Charting: change chart of accounts,
    - Posting: change account balances,
    - Closing: compound period end operation on a ledger.
    """


class Charting(Operation):
    """Change chart of accounts."""


class Posting(Operation):
    """Change account balances."""


class Closing(Operation):
    """Close accounts at period end."""


class AbacusError(Exception):
    pass

    @staticmethod
    def must_not_exist(collection: Iterable[str], name: str):
        if name in collection:
            raise AbacusError(f"Account {name} already exists.")

    @staticmethod
    def must_exist(collection: Iterable[str], name: str):
        if name not in collection:
            raise AbacusError(f"Account {name} not found.")


class T5(Enum):
    Asset = "asset"
    Liability = "liability"
    Equity = "equity"
    Income = "income"
    Expense = "expense"

    def __repr__(self):
        return self.value.capitalize()


class SaveLoadMixin:
    """A mix-in class for loading and saving pydantic models to files."""

    @classmethod
    def load(cls, filename: str | Path):
        return cls.model_validate_json(Path(filename).read_text())  # type: ignore

    def save(self, filename: str | Path, allow_overwrite: bool = False):
        if not allow_overwrite and Path(filename).exists():
            raise FileExistsError(f"File already exists: {filename}")
        content = self.model_dump_json(indent=2, warnings=False)  # type: ignore
        Path(filename).write_text(content)  # type: ignore
